@@ -13,7 +13,6 @@ mod aat_layout_trak_table;
 mod aat_map;
 pub mod common;
 pub mod face;
-mod glyph_set;
 mod kerning;
 mod machine_cursor;
 mod ot;
@@ -49,7 +48,10 @@ mod ot_shaper_use_table;
 mod aat_layout_common;
 mod ot_shaper_vowel_constraints;
 mod paint_extents;
+mod set_digest;
 pub mod shape;
+#[cfg(feature = "wasm-shaper")]
+mod shape_wasm;
 mod tag;
 mod tag_table;
 mod text_parser;
